/*

  ci_incrbar.c: Counter Increments using threads synchronized with the Sense Reversing Centralized Incremental Barrier.
  Copyright (C) 2001 Thomas Repantis <darkzero@otenet.gr>
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  
*/
 
/*
  
  This program increments counters using threads synchronized with the Sense Reversing Centralized Incremental Barrier.

*/

#include <stdio.h>
#include <stdlib.h>   /* To use strtoul(). */
#include <errno.h>    /* To use errno. */
#include <sys/time.h> /* To use gettimeofday(). */
#include <thread.h>
#include "libsync.h"  /* To use the libsync library we implemented. */


#ifdef MEASURE_TIME

#define CALLS_TO_GETTIMEOFDAY 2
#define MICROS_PER_SECOND 1000000

#endif

/* These are variables to be used by all functions (global), in this 
   file only (static) and changeable by many threads (volatile). */

static volatile long int counter[MAX_THR_NUM];
static unsigned long int target_value = 0;
static int threads_num = 0;
struct incr_barrier thr_incr_barrier;

#ifdef MEASURE_TIME
/* We use padding to the size of the cache line to avoid false sharing.
   Therefore we put all time variables and structs inside a t(ime) struct. */

struct {
/* In this array we keep the timeval structures returned by each call to gettimeofday(). 
   Basically total_times[2] should be enough for the total execution time.               */
struct timeval _pad12[TIMEVALPADS_PER_CACHE_LINE];
struct timeval total_times[CALLS_TO_GETTIMEOFDAY][TIMEVALPADS_PER_CACHE_LINE - 1];
/* In this array we keep -for each thread- the timeval structures returned by each call to gettimeofday(). */
struct timeval times[MAX_THR_NUM][CALLS_TO_GETTIMEOFDAY][TIMEVALPADS_PER_CACHE_LINE - 2];
unsigned long int time_elapsed;
long _pad13[LONGPADS_PER_CACHE_LINE - 1];
unsigned long int time_till_bar_leave[MAX_THR_NUM][LONGPADS_PER_CACHE_LINE - 1];
} t;
#endif

/* This is the function executed by threads after they are created. */
void *startfunc(void *arg) {
  
  long int i;

  for (i = 0; i < target_value; i++)
    {
      counter[(int)arg]++;
      
#ifdef MEASURE_TIME
      gettimeofday(&t.times[(int)arg][0][0], NULL);
#endif

      incr_barrier(&thr_incr_barrier);

#ifdef MEASURE_TIME
      gettimeofday(&t.times[(int)arg][1][0], NULL);
      t.time_till_bar_leave[(int)arg][0] += (t.times[(int)arg][1][0].tv_sec - t.times[(int)arg][0][0].tv_sec)*MICROS_PER_SECOND;
      t.time_till_bar_leave[(int)arg][0] += (t.times[(int)arg][1][0].tv_usec - t.times[(int)arg][0][0].tv_usec);   
#endif
      
#ifndef MEASURE_TIME
      /* Present the counter values -only when no measurements are taking place to avoid the printing delay. */ 
      fprintf(stdout, "Thread %d has reached counter value %d.\n", (int)arg, counter[(int)arg]);
      fflush(stdout);
#endif
    }
}


int main(int argc, char *argv[])
{

  thread_t tid;
  int i;
  
  /* Introduce yourself. */

  printf("\n");
  printf("ci_incrbar.c: Counter Increments using threads synchronized with the Sense Reversing Centralized Incremental Barrier.\n");  
  printf("Copyright (C) 2001 Thomas Repantis <darkzero@otenet.gr>\n");
  printf("This program comes with ABSOLUTELY NO WARRANTY. This is free software, and you are welcome to redistribute it under certain conditions. ci_incrbar is licensed under the GNU General Public License. See the file COPYING for details.\n");
  printf("\n");

  if( argc != 3)
    {
      printf("Usage: ci_incrbar [Number of threads to be used] [Target value for the counters to reach.]\n");
      printf("For example: ci_incrbar 4 1000000\n");
      printf("\n");
      exit(1);
    }
  
  
  /* Convert the command line argument strings to unsigned long integers. */
  
  errno = 0;
  threads_num = (int)strtoul(argv[1], (char**)NULL, 10);
  if( errno != 0 )
    {
      fprintf(stderr, "Error on the number of threads (%s) given.\n", argv[1]);
      fflush(stderr);
      exit(1);
    }

  errno = 0;
  target_value = strtoul(argv[2], (char**)NULL, 10);
  if( errno != 0 )
    {
      fprintf(stderr, "Error on the target value (%s) given.\n", argv[2]);
      fflush(stderr);
      exit(1);
    }
  
  fprintf(stdout, "The number of threads given is interpreted as: %d\n", threads_num);
  fflush(stdout);
  
  fprintf(stdout, "The target value given is interpreted as: %lu\n", target_value);
  fflush(stdout);
  
  /* Check for valid input. */

  if( threads_num == 0)
    {
      fprintf(stderr, "Error on the number of threads given (%d). \n", threads_num);
      fprintf(stderr, "There must at least be one thread.\n");
      fflush(stderr);
      exit(1);
    }


  if( threads_num > MAX_THR_NUM)
    {
      fprintf(stderr, "Error on the number of threads given (%d). \n", threads_num);
      fprintf(stderr, "Maximum number of threads allowed is %d.\n", MAX_THR_NUM);
      fflush(stderr);
      exit(1);
    }


  /* Initialize the counter array. */
  memset(&counter, 0, threads_num);
  /* Using memset we initialize the counter array faster than with a for loop. */

  /* Initialize the incr_bar. */
  incr_barrier_init(&thr_incr_barrier, threads_num);
  /* We pass as an argument the number of threads taking part in the barrier too, 
     allowing the possibility that not all threads take part in all barriers.     */

#ifdef MEASURE_TIME
  gettimeofday(&t.total_times[0][0], NULL);
#endif

  /* Create the threads. */
  for (i=0; i<threads_num; i++) {
    /* Create one thread with default stack and stack size *
     * and make it execute startfunc. The argument of      *
     * startfunc is an "id" of the thread. The thread is   *
     * bound (this is very important - the thread is       *
     * scheduled on a system-wide basis). The actual id of * 
     * the thread is returned to tid.                      */
    thr_create(NULL, NULL, startfunc, (void *)i, THR_BOUND, &tid);
    fprintf(stdout, "Created thread %d with ID %d.\n", i, tid);
    fflush(stdout);
  }
  
  /* Now wait for the threads created earlier to finish */
  for (i=0; i<threads_num; i++) {
    thr_join(NULL, NULL, NULL);
  }	

#ifdef MEASURE_TIME
  gettimeofday(&t.total_times[1][0], NULL);
  t.time_elapsed = (t.total_times[1][0].tv_sec - t.total_times[0][0].tv_sec)*MICROS_PER_SECOND;
  t.time_elapsed += (t.total_times[1][0].tv_usec - t.total_times[0][0].tv_usec);   
#endif


  /* Present the counter final values -only when time measurements are taking place 
     and therefore the values haven't already been presented. */
#ifdef MEASURE_TIME
  for (i=0; i<threads_num; i++)
    {
      fprintf(stdout, "Thread %d has reached counter value %lu.\n", i, counter[i]);
      fflush(stdout);
    }
#endif
  
  /* Present the time measurements. */ 
#ifdef MEASURE_TIME
  /* Present the measured times in microsecond precision. */
  /* And do not be alarmed if the total time is smaller than the sum of the times for the several threads. 
     You see, some of the times of the several threads are spent parallelly.                               */
  fprintf(stdout, "Measured time to increment the counter using threads synchronized with the Sense Reversing Centralized Incremental Barrier: %lu microseconds (%f seconds)\n", t.time_elapsed, t.time_elapsed/((double)MICROS_PER_SECOND));
  for (i=0; i<threads_num; i++)
    {
      fprintf(stdout, "Measured time spent waiting until the barrier has been reached for thread %d: %lu microseconds (%f seconds)\n", i, t.time_till_bar_leave[i][0], t.time_till_bar_leave[i][0]/((double)MICROS_PER_SECOND));
      fflush(stdout);
    }
#endif

}

	


